# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import glob
import os
import sys
import textwrap
import re

import numpy as np
import matplotlib
import mpl_toolkits.mplot3d
import errno
from numpy.testing import (
    assert_almost_equal,
    assert_equal,
)

import MDAnalysis as mda
from MDAnalysis.analysis import hole2
from MDAnalysis.analysis.hole2.utils import check_and_fix_long_filename
from MDAnalysis.exceptions import ApplicationError
from MDAnalysisTests.datafiles import PDB_HOLE, MULTIPDB_HOLE, DCD
from MDAnalysisTests import executable_not_found
from MDAnalysis.analysis.hole2.templates import exe_err

def rlimits_missing():
    # return True if resources module not accesible (ie setting of rlimits)
    try:
        # on Unix we can manipulate our limits: http://docs.python.org/2/library/resource.html
        import resource

        soft_max_open_files, hard_max_open_files = resource.getrlimit(
            resource.RLIMIT_NOFILE)
    except ImportError:
        return True
    return False


class TestCheckAndFixLongFilename(object):

    max_length = 70
    filename = 'a'*(max_length-4) + '.pdb'

    def test_short(self):
        fixed = check_and_fix_long_filename(self.filename)
        assert self.filename == fixed

    def test_relative(self):
        abspath = os.path.abspath(self.filename)
        if len(abspath) > self.max_length:
            fixed = check_and_fix_long_filename(abspath)
            assert fixed == self.filename

    @pytest.mark.skipif(os.name == 'nt' and sys.maxsize <= 2**32,
                        reason="FileNotFoundError on Win 32-bit")
    def test_symlink_dir(self, tmpdir):
        dirname = 'really_'*20 +'long_name'
        short_name = self.filename[-20:]
        path = os.path.join(dirname, short_name)
        with tmpdir.as_cwd():
            os.makedirs(dirname)
            u = mda.Universe(PDB_HOLE)
            u.atoms.write(path)

            fixed = check_and_fix_long_filename(path)
            assert os.path.islink(fixed)
            assert fixed.endswith(short_name)

    @pytest.mark.skipif(os.name == 'nt' and sys.maxsize <= 2**32,
                        reason="OSError: symbolic link privilege not held")
    def test_symlink_file(self, tmpdir):
        long_name = 'a'*10 + self.filename

        with tmpdir.as_cwd():
            fixed = check_and_fix_long_filename(long_name)
            assert os.path.islink(fixed)
            assert not fixed.endswith(long_name)


@pytest.mark.skipif(executable_not_found("hole"),
                    reason="Test skipped because HOLE not found")
class TestHole(object):
    filename = PDB_HOLE
    random_seed = 31415
    profile_length = 425
    rxn_coord_mean = -1.41225
    radius_min = 1.19707

    def test_correct_input(self, tmpdir):
        with tmpdir.as_cwd():
            hole2.hole(self.filename, random_seed=self.random_seed,
                       infile='hole.inp')

        infile = str(tmpdir.join('hole.inp'))
        with open(infile, 'r') as f:
            contents = f.read()

        hole_input = textwrap.dedent("""
            RADIUS simple2.rad
            SPHPDB hole.sph
            SAMPLE 0.200000
            ENDRAD 22.000000
            IGNORE SOL WAT TIP HOH K   NA  CL 
            SHORTO 0
            RASEED 31415
            """)

        # don't check filenames
        assert contents.endswith(hole_input)

    def test_input_options(self, tmpdir):
        u = mda.Universe(PDB_HOLE)
        cog = u.select_atoms('protein').center_of_geometry()

        with tmpdir.as_cwd():
            hole2.hole(self.filename, random_seed=self.random_seed,
                       infile='hole.inp', cpoint=cog,
                       ignore_residues=[])

        infile = str(tmpdir.join('hole.inp'))
        with open(infile, 'r') as f:
            contents = f.read()

        hole_input = textwrap.dedent("""
            RADIUS simple2.rad
            SPHPDB hole.sph
            SAMPLE 0.200000
            ENDRAD 22.000000
            IGNORE 
            SHORTO 0
            RASEED 31415
            CPOINT -0.0180961507 -0.0122730583 4.1497999943
            """)

        # don't check filenames
        assert contents.endswith(hole_input)

    def test_correct_profile_values(self, tmpdir):
        with tmpdir.as_cwd():
            profiles = hole2.hole(self.filename, random_seed=self.random_seed)

        values = list(profiles.values())
        assert_equal(len(values), 1,
                     err_msg='Should only have 1 HOLE profile')
        profile = values[0]
        assert_equal(len(profile), self.profile_length,
                     err_msg='Wrong number of points in HOLE profile')
        assert_almost_equal(profile.rxn_coord.mean(),
                            self.rxn_coord_mean,
                            err_msg='Wrong mean HOLE rxn_coord')
        assert_almost_equal(profile.radius.min(),
                            self.radius_min,
                            err_msg='Wrong minimum HOLE radius')

    # HOLE cannot read NAMD CHARMM files as is written by MDAnalysis
    # fails with Linux HOLE?
    # def test_dcd(self, tmpdir):
    #     with tmpdir.as_cwd():
    #         u = mda.Universe(PSF, DCD)
    #         n_frames = len(u.trajectory)
    #         keep_frames = 12
    #         step = 3
    #         skip = n_frames-keep_frames
    #         filename = 'temp.pdb'
    #         u.atoms.write(filename)

    #         profiles = hole2.hole(filename, random_seed=self.random_seed, dcd=DCD,
    #                               dcd_iniskip=skip,
    #                               dcd_step=step, infile='hole.inp')
    #         with open('hole.inp', 'r') as f:
    #             contents = f.read()
    #         assert contents.endswith('CHARMS {} {}\n'.format(skip, step-1))

    #     assert_equal(len(profiles), int(keep_frames/step))

    def test_application_error(self, tmpdir):
        with tmpdir.as_cwd():
            with pytest.raises(ApplicationError):
                hole2.hole(self.filename, dcd=DCD)

    def test_output_level(self, tmpdir):
        with tmpdir.as_cwd():
            with pytest.warns(UserWarning, match="needs to be < 3"):
                profiles = hole2.hole(self.filename,
                                      random_seed=self.random_seed,
                                      output_level=100)
            assert len(profiles) == 0

    def test_keep_files(self, tmpdir):
        with tmpdir.as_cwd():
            hole2.hole(self.filename, random_seed=self.random_seed,
                       infile='hole.inp',
                       keep_files=False)
            sphpdbs = tmpdir.join('*.sph')
            assert len(glob.glob(str(sphpdbs))) == 0
            outfiles = tmpdir.join('*.out')
            assert len(glob.glob(str(outfiles))) == 0
            vdwradii = tmpdir.join('simple2.rad')
            assert len(glob.glob(str(vdwradii))) == 0
            pdbfiles = tmpdir.join('*.pdb')
            assert len(glob.glob(str(pdbfiles))) == 0
            oldfiles = tmpdir.join('*.old')
            assert len(glob.glob(str(oldfiles))) == 0


@pytest.mark.skipif(executable_not_found("hole"),
                    reason="Test skipped because HOLE not found")
class BaseTestHole(object):
    filename = MULTIPDB_HOLE
    start = 5
    stop = 7
    random_seed = 31415

    # HOLE is so slow so we only run it once and keep it in
    # the class; note that you may not change universe.trajectory
    # (eg iteration) because this is not safe in parallel
    @pytest.fixture()
    def universe(self):
        return mda.Universe(self.filename)

    @pytest.fixture()
    def hole(self, universe, tmpdir):
        with tmpdir.as_cwd():
            h = hole2.HoleAnalysis(universe)
            h.run(start=self.start, stop=self.stop,
                  random_seed=self.random_seed)
        return h

    @pytest.fixture()
    def frames(self, universe):
        return [ts.frame for ts in universe.trajectory[self.start:self.stop]]

    @pytest.fixture()
    def profiles(self, hole, frames):
        return [hole.results.profiles[f] for f in frames]

    @pytest.mark.parametrize("attrname", ["sphpdbs", "outfiles", "profiles"])
    def test_deprecated_warning(self, hole, attrname):
        wmsg = (f"The `{attrname}` attribute was deprecated in "
                "MDAnalysis 2.0.0 and will be removed in MDAnalysis 3.0.0. "
                f"Please use `results.{attrname}` instead.")
        with pytest.warns(DeprecationWarning, match=wmsg):
            value = getattr(hole, attrname)
            assert value is hole.results[attrname]


@pytest.mark.skipif(executable_not_found('hole'),
                    reason="Test skipped because HOLE not found")
class TestOSError:

    @pytest.fixture()
    def universe(self):
        return mda.Universe(MULTIPDB_HOLE)

    def test_hole_method_oserror(self):
        errmsg = exe_err.format(name='dummy_path', kw='executable')
        with pytest.raises(OSError, match=errmsg):
            h = hole2.hole(PDB_HOLE, executable='dummy_path')

    def test_hole_oserror(self, universe):
        errmsg = exe_err.format(name='dummy_path', kw='executable')
        with pytest.raises(OSError, match=errmsg):
            h = hole2.HoleAnalysis(universe, executable='dummy_path')

    def test_sos_triangle_oserror(self, universe):
        errmsg = exe_err.format(name='dummy_path', kw='sos_triangle')
        with pytest.raises(OSError, match=errmsg):
            h = hole2.HoleAnalysis(universe, sos_triangle='dummy_path')

    def test_sph_process_oserror(self, universe):
        errmsg = exe_err.format(name='dummy_path', kw='sph_process')
        with pytest.raises(OSError, match=errmsg):
            h = hole2.HoleAnalysis(universe, sph_process='dummy_path')


class TestHoleAnalysis(BaseTestHole):

    def test_correct_profile_values(self, hole, frames):
        assert_equal(sorted(hole.results.profiles.keys()), frames,
                     err_msg="hole.results.profiles.keys() should contain the frame numbers")
        assert_equal(list(hole.frames), frames,
                     err_msg="hole.frames should contain the frame numbers")
        data = np.transpose([(len(p), p.rxn_coord.mean(), p.radius.min())
                             for p in hole.results.profiles.values()])
        assert_equal(data[0], [401, 399], err_msg="incorrect profile lengths")
        assert_almost_equal(data[1], [1.98767,  0.0878],
                            err_msg="wrong mean HOLE rxn_coord")
        assert_almost_equal(data[2], [1.19819, 1.29628],
                            err_msg="wrong minimum radius")

    def test_min_radius(self, hole):
        values = np.array([[5., 1.19819],
                           [6., 1.29628]])
        assert_almost_equal(hole.min_radius(), values,
                            err_msg="min_radius() array not correct")

    def test_context_manager(self, universe, tmpdir):
        with tmpdir.as_cwd():
            with hole2.HoleAnalysis(universe) as h:
                h.run()
                h.run()  # run again for *.old files
                h.create_vmd_surface(filename='hole.vmd')

        sphpdbs = tmpdir.join('*.sph')
        assert len(glob.glob(str(sphpdbs))) == 0
        outfiles = tmpdir.join('*.out')
        assert len(glob.glob(str(outfiles))) == 0
        vdwradii = tmpdir.join('simple2.rad')
        assert len(glob.glob(str(vdwradii))) == 0
        pdbfiles = tmpdir.join('*.pdb')
        assert len(glob.glob(str(pdbfiles))) == 0
        oldfiles = tmpdir.join('*.old')
        assert len(glob.glob(str(oldfiles))) == 0
        vmd_file = tmpdir.join('hole.vmd')
        assert len(glob.glob(str(vmd_file))) == 1

    @pytest.mark.parametrize("start,stop,step", [
        (1, 9, 2), (1, None, 3), (5, -2, None)])
    def test_nonzero_start_surface(self, universe, tmpdir,
                                   start, stop, step,
                                   surface="hole.vmd"):
        # Issue 3476
        with tmpdir.as_cwd():
            h = hole2.HoleAnalysis(universe)
            h.run(start=start, stop=stop, step=step)
            h.create_vmd_surface(filename=surface)

            found_frame_indices = []
            with open(surface) as s:
                for line in s:
                    m = re.match(r"set triangles\((?P<frame>\d+)\)", line)
                    if m:
                        found_frame_indices.append(m.group('frame'))
            found_frame_indices = np.array(found_frame_indices, dtype=int)
        assert_equal(found_frame_indices,
                     np.arange(len(universe.trajectory[start:stop:step])),
                     err_msg="wrong frame indices in VMD surface file")

    def test_output_level(self, tmpdir, universe):
        with tmpdir.as_cwd():
            with pytest.warns(UserWarning, match='needs to be < 3'):
                h = hole2.HoleAnalysis(universe,
                                       output_level=100)
                h.run(start=self.start,
                      stop=self.stop, random_seed=self.random_seed)

            # no profiles
            assert len(h.results.profiles) == 0

    def test_cpoint_geometry(self, tmpdir, universe):
        protein = universe.select_atoms('protein')
        cogs = [protein.center_of_geometry() for ts in universe.trajectory]
        with tmpdir.as_cwd():
            h = hole2.HoleAnalysis(universe,
                                   select='protein',
                                   cpoint='center_of_geometry',
                                   write_input_files=True)
            h.run(start=self.start,
                  stop=self.stop, random_seed=self.random_seed)

            infiles = sorted(glob.glob(str(tmpdir.join('hole*.inp'))))
            for file, cog in zip(infiles, cogs[self.start:self.stop]):
                with open(file, 'r') as f:
                    line = f.read().split('CPOINT')[1].split('\n')[0]
                arr = np.array(list(map(float, line.split())))
                assert_almost_equal(arr, cog)

    # plotting
    def test_plot(self, hole, frames, profiles):
        ax = hole.plot(label=True, frames=None, y_shift=1)
        err_msg = "HoleAnalysis.plot() did not produce an Axes instance"
        assert isinstance(ax, matplotlib.axes.Axes), err_msg
        lines = ax.get_lines()[:]
        assert len(lines) == hole.n_frames
        for i, (line, frame, profile) in enumerate(zip(lines, frames, profiles)):
            x, y = line.get_data()
            assert_almost_equal(x, profile.rxn_coord)
            assert_almost_equal(y, profile.radius + i)
            assert line.get_label() == str(frame)

    def test_plot_mean_profile(self, hole, frames, profiles):
        binned, bins = hole.bin_radii(bins=100)
        mean = np.array(list(map(np.mean, binned)))
        stds = np.array(list(map(np.std, binned)))
        midpoints = 0.5 * bins[1:] + bins[:-1]
        ylow = list(mean-(2*stds))
        yhigh = list(mean+(2*stds))

        ax = hole.plot_mean_profile(bins=100, n_std=2)

        # test fillbetween standard deviation
        children = ax.get_children()
        poly = []
        for x in children:
            if isinstance(x, matplotlib.collections.PolyCollection):
                poly.append(x)
        assert len(poly) == 1
        xp, yp = poly[0].get_paths()[0].vertices.T
        assert_almost_equal(np.unique(xp), np.unique(midpoints))
        assert_almost_equal(np.unique(yp), np.unique(ylow+yhigh))

        # test mean line
        lines = ax.get_lines()
        assert len(lines) == 1
        xl, yl = lines[0].get_data()
        assert_almost_equal(xl, midpoints)
        assert_almost_equal(yl, mean)

    @pytest.mark.skipif(sys.version_info < (3, 1),
                        reason="get_data_3d requires 3.1 or higher")
    def test_plot3D(self, hole, frames, profiles):
        ax = hole.plot3D(frames=None, r_max=None)
        err_msg = "HoleAnalysis.plot3D() did not produce an Axes3D instance"
        assert isinstance(ax, mpl_toolkits.mplot3d.Axes3D), err_msg
        lines = ax.get_lines()[:]
        assert len(lines) == hole.n_frames

        for line, frame, profile in zip(lines, frames, profiles):
            x, y, z = line.get_data_3d()
            assert_almost_equal(x, profile.rxn_coord)
            assert_almost_equal(np.unique(y), [frame])
            assert_almost_equal(z, profile.radius)
            assert line.get_label() == str(frame)

    @pytest.mark.skipif(sys.version_info < (3, 1),
                        reason="get_data_3d requires 3.1 or higher")
    def test_plot3D_rmax(self, hole, frames, profiles):
        ax = hole.plot3D(r_max=2.5)
        err_msg = "HoleAnalysis.plot3D(rmax=float) did not produce an Axes3D instance"
        assert isinstance(ax, mpl_toolkits.mplot3d.Axes3D), err_msg

        lines = ax.get_lines()[:]

        for line, frame, profile in zip(lines, frames, profiles):
            x, y, z = line.get_data_3d()
            assert_almost_equal(x, profile.rxn_coord)
            assert_almost_equal(np.unique(y), [frame])
            radius = np.where(profile.radius > 2.5, np.nan, profile.radius)
            assert_almost_equal(z, radius)
            assert line.get_label() == str(frame)

    def test_none_filename(self, tmpdir):
        universe_none_filename = mda.Universe(PDB_HOLE, in_memory=True)
        universe_none_filename.trajectory.filename = None
        with tmpdir.as_cwd():
            with hole2.HoleAnalysis(universe_none_filename) as h:
                h.run()


class TestHoleAnalysisLong(BaseTestHole):

    start = 0
    stop = 11

    rmsd = np.array([6.10501252e+00, 4.88398472e+00, 3.66303524e+00, 2.44202454e+00,
                     1.22100521e+00, 1.67285541e-07, 1.22100162e+00, 2.44202456e+00,
                     3.66303410e+00, 4.88398478e+00, 6.10502262e+00])

    @pytest.fixture
    def order_parameter_keys_values(self, hole):
        op = hole.over_order_parameters(self.rmsd, frames=None)
        return op.keys(), op.values()

    def test_gather(self, hole):
        gd = hole.gather(flat=False)
        for i, p in enumerate(hole.results.profiles.values()):
            assert_almost_equal(p.rxn_coord, gd['rxn_coord'][i])
            assert_almost_equal(p.radius, gd['radius'][i])
            assert_almost_equal(p.cen_line_D, gd['cen_line_D'][i])

    def test_gather_flat(self, hole):
        gd = hole.gather(flat=True)
        i = 0
        for p in hole.results.profiles.values():
            j = i+len(p.rxn_coord)
            assert_almost_equal(p.rxn_coord, gd['rxn_coord'][i:j])
            assert_almost_equal(p.radius, gd['radius'][i:j])
            assert_almost_equal(p.cen_line_D, gd['cen_line_D'][i:j])
            i = j
        assert_equal(i, len(gd['rxn_coord']))

    def test_min_radius(self, hole):
        rad = hole.min_radius()
        for (f1, p), (f2, r) in zip(hole.results.profiles.items(), rad):
            assert_equal(f1, f2)
            assert_almost_equal(min(p.radius), r)

    def test_over_order_parameters(self, hole):
        op = self.rmsd
        profiles = hole.over_order_parameters(op, frames=None)
        assert len(op) == len(profiles)

        for key, rmsd in zip(profiles.keys(), np.sort(op)):
            assert key == rmsd

        idx = np.argsort(op)
        arr = np.array(list(hole.results.profiles.values()), dtype=object)

        for op_prof, arr_prof in zip(profiles.values(), arr[idx]):
            assert op_prof is arr_prof

    def test_over_order_parameters_file(self, hole, tmpdir):
        op = self.rmsd
        with tmpdir.as_cwd():
            np.savetxt('rmsd.dat', self.rmsd)
            profiles = hole.over_order_parameters('rmsd.dat', frames=None)

        assert len(op) == len(profiles)

        for key, rmsd in zip(profiles.keys(), np.sort(op)):
            assert key == rmsd

        idx = np.argsort(op)
        arr = np.array(list(hole.results.profiles.values()), dtype=object)

        for op_prof, arr_prof in zip(profiles.values(), arr[idx]):
            assert op_prof is arr_prof

    def test_over_order_parameters_missing_file(self, hole):
        with pytest.raises(ValueError) as exc:
            hole.over_order_parameters('missing.dat')
        assert 'not found' in str(exc.value)

    def test_over_order_parameters_invalid_file(self, hole):
        with pytest.raises(ValueError) as exc:
            hole.over_order_parameters(PDB_HOLE)
        assert 'Could not parse' in str(exc.value)

    def test_over_order_parameters_frames(self, hole):
        op = self.rmsd
        n_frames = 7
        profiles = hole.over_order_parameters(op, frames=np.arange(n_frames))
        assert len(profiles) == n_frames
        for key, rmsd in zip(profiles.keys(), np.sort(op[:n_frames])):
            assert key == rmsd

        idx = np.argsort(op[:n_frames])
        values = list(hole.results.profiles.values())[:n_frames]

        arr = np.array(values, dtype=object)
        for op_prof, arr_prof in zip(profiles.values(), arr[idx]):
            assert op_prof is arr_prof

    def test_bin_radii(self, hole):
        radii, bins = hole.bin_radii(bins=100)
        dct = hole.gather(flat=True)
        coords = dct['rxn_coord']

        assert len(bins) == 101
        assert_almost_equal(bins[0], coords.min())
        assert_almost_equal(bins[-1], coords.max())
        assert len(radii) == (len(bins)-1)

        # check first frame profile
        first = hole.results.profiles[0]
        for row in first:
            coord = row.rxn_coord
            rad = row.radius
            for i, (lower, upper) in enumerate(zip(bins[:-1], bins[1:])):
                if coord > lower and coord <= upper:
                    assert rad in radii[i]
                    break
            else:
                raise AssertionError('Radius not in binned radii')

    @pytest.mark.parametrize('midpoint', [1.5, 1.8, 2.0, 2.5])
    def test_bin_radii_range(self, hole, midpoint):
        radii, bins = hole.bin_radii(bins=100,
                                     range=(midpoint, midpoint))
        dct = hole.gather(flat=True)
        coords = dct['rxn_coord']

        assert len(bins) == 101
        low = midpoint - 0.5
        high = midpoint + 0.5
        assert_almost_equal(bins[0], low)
        assert_almost_equal(bins[-1], high)
        assert len(radii) == (len(bins)-1)

        # check first frame profile
        first = hole.results.profiles[0]
        for row in first:
            coord = row.rxn_coord
            rad = row.radius
            if coord > low and coord <= high:
                for i, (lower, upper) in enumerate(zip(bins[:-1], bins[1:])):
                    if coord > lower and coord <= upper:
                        assert rad in radii[i]
                        break
                else:
                    raise AssertionError('Radius not in binned radii')
            else:
                assert not any([rad in x for x in radii])

    def test_bin_radii_edges(self, hole):
        brange = list(np.linspace(1.0, 2.0, num=101, endpoint=True))
        moved = brange[30:] + brange[10:30] + brange[:10]
        e_radii, e_bins = hole.bin_radii(bins=moved, range=(0.0, 0.0))
        r_radii, r_bins = hole.bin_radii(bins=100, range=(1.5, 1.5))
        assert_almost_equal(e_bins, r_bins)
        for e, r in zip(e_radii, r_radii):
            assert_almost_equal(e, r)

    def test_histogram_radii(self, hole):
        means, _ = hole.histogram_radii(aggregator=np.mean,
                                        bins=100)
        radii, _ = hole.bin_radii(bins=100)
        assert means.shape == (100,)
        for r, m in zip(radii, means):
            assert_almost_equal(r.mean(), m)

    # plotting

    def test_plot_select_frames(self, hole, frames, profiles):
        ax = hole.plot(label=True, frames=[2, 3], y_shift=1)
        err_msg = "HoleAnalysis.plot() did not produce an Axes instance"
        assert isinstance(ax, matplotlib.axes.Axes), err_msg
        lines = ax.get_lines()[:]
        assert len(lines) == 2
        for i, (line, frame, profile) in enumerate(zip(lines, frames[2:4], profiles[2:4])):
            x, y = line.get_data()
            assert_almost_equal(x, profile.rxn_coord)
            assert_almost_equal(y, profile.radius + i)
            assert line.get_label() == str(frame)

    @pytest.mark.parametrize('agg', [np.max, np.mean, np.std, np.min])
    def test_plot_order_parameters(self, hole, order_parameter_keys_values,
                                   agg):
        opx = np.array(list(order_parameter_keys_values[0]))
        opy = np.array([agg(p.radius) for p in order_parameter_keys_values[1]])
        ax = hole.plot_order_parameters(self.rmsd, aggregator=agg, frames=None)
        err_msg = ("HoleAnalysis.plot_order_parameters()"
                   "did not produce an Axes instance")
        assert isinstance(ax, matplotlib.axes.Axes), err_msg

        lines = ax.get_lines()
        assert len(lines) == 1
        x, y = lines[0].get_data()
        assert_almost_equal(x, opx)
        assert_almost_equal(y, opy)

    @pytest.mark.skipif(sys.version_info < (3, 1),
                        reason="get_data_3d requires 3.1 or higher")
    def test_plot3D_order_parameters(self, hole, order_parameter_keys_values):
        opx = np.array(list(order_parameter_keys_values[0]))
        profiles = np.array(list(order_parameter_keys_values[1]))
        ax = hole.plot3D_order_parameters(self.rmsd, frames=None)
        err_msg = ("HoleAnalysis.plot3D_order_parameters() "
                   "did not produce an Axes3D instance")
        assert isinstance(ax, mpl_toolkits.mplot3d.Axes3D), err_msg

        lines = ax.get_lines()
        assert len(lines) == hole.n_frames
        for line, opx_, profile in zip(lines, opx, profiles):
            x, y, z = line.get_data_3d()
            assert_almost_equal(x, profile.rxn_coord)
            assert_almost_equal(np.unique(y), np.array([opx_]))
            assert_almost_equal(z, profile.radius)

    @pytest.mark.skipif(sys.version_info > (3, 1),
                        reason="get_data_3d requires 3.1 or higher")
    def test_plot3D_order_parameters(self, hole, order_parameter_keys_values):
        opx = np.array(list(order_parameter_keys_values[0]))
        profiles = np.array(list(order_parameter_keys_values[1]))
        ax = hole.plot3D_order_parameters(self.rmsd, frames=None)
        err_msg = ("HoleAnalysis.plot3D_order_parameters() "
                   "did not produce an Axes3D instance")
        assert isinstance(ax, mpl_toolkits.mplot3d.Axes3D), err_msg

        lines = ax.get_lines()
        assert len(lines) == hole.n_frames
        for line, opx_, profile in zip(lines, opx, profiles):
            x, y = line.get_data()
            assert_almost_equal(x, profile.rxn_coord)
            assert_almost_equal(np.unique(y), np.array([opx_]))

@pytest.mark.skipif(executable_not_found("hole"),
                    reason="Test skipped because HOLE not found")
class TestHoleModule(object):
    try:
        # on Unix we can manipulate our limits: http://docs.python.org/2/library/resource.html
        import resource
        soft_max_open_files, hard_max_open_files = resource.getrlimit(
            resource.RLIMIT_NOFILE)
    except ImportError:
        pass

    @staticmethod
    @pytest.fixture()
    def universe():
        return mda.Universe(MULTIPDB_HOLE)

    @pytest.mark.skipif(rlimits_missing,
                        reason="Test skipped because platform does not allow setting rlimits")
    def test_hole_module_fd_closure(self, universe, tmpdir):
        """test open file descriptors are closed (MDAnalysisTests.analysis.test_hole.TestHoleModule): Issue 129"""
        # If Issue 129 isn't resolved, this function will produce an OSError on
        # the system, and cause many other tests to fail as well.
        #
        # Successful test takes ~10 s, failure ~2 s.

        # Hasten failure by setting "ulimit -n 64" (can't go too low because of open modules etc...)
        import resource

        # ----- temporary hack -----
        # on Mac OS X (on Travis) we run out of open file descriptors
        # before even starting this test (see
        # https://github.com/MDAnalysis/mdanalysis/pull/901#issuecomment-231938093);
        # if this issue is solved by #363 then revert the following
        # hack:
        #
        import platform
        if platform.platform() == "Darwin":
            max_open_files = 512
        else:
            max_open_files = 64
        #
        # --------------------------

        resource.setrlimit(resource.RLIMIT_NOFILE,
                           (max_open_files, self.hard_max_open_files))

        with tmpdir.as_cwd():
            try:
                H = hole2.HoleAnalysis(universe, cvect=[0, 1, 0], sample=20.0)
            finally:
                self._restore_rlimits()

            # pretty unlikely that the code will get through 2 rounds if the MDA
            # issue 129 isn't fixed, although this depends on the file descriptor
            # open limit for the machine in question
            try:
                for i in range(2):
                    # will typically get an OSError for too many files being open after
                    # about 2 seconds if issue 129 isn't resolved
                    H.run()
            except OSError as err:
                if err.errno == errno.EMFILE:
                    raise pytest.fail(
                        "hole2.HoleAnalysis does not close file descriptors (Issue 129)")
                raise
            finally:
                # make sure to restore open file limit !!
                self._restore_rlimits()

    def _restore_rlimits(self):
        try:
            import resource
            resource.setrlimit(resource.RLIMIT_NOFILE,
                               (self.soft_max_open_files, self.hard_max_open_files))
        except ImportError:
            pass



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""Tests for the API definition of Timestep


_TestTimestepInterface tests the Readers are correctly using Timesteps
"""
import itertools
import numpy as np
from numpy.testing import assert_equal, assert_allclose, assert_array_almost_equal

from MDAnalysis.lib.mdamath import triclinic_vectors
import MDAnalysis as mda
from MDAnalysisTests.datafiles import (PSF, XYZ_five, INPCRD, DCD, DLP_CONFIG,
                                       DLP_HISTORY, DMS, GMS_ASYMOPT, GRO, XTC,
                                       TRR, LAMMPSdata, LAMMPSdata2,
                                       LAMMPSdcd2, mol2_molecules, PDB_small,
                                       PDBQT_input, PQR, PRM, TRJ, PRMncdf,
                                       NCDF, TRZ_psf, TRZ)

from MDAnalysisTests.coordinates.base import assert_timestep_equal, assert_timestep_almost_equal
import pytest


class TestTimestep(object):
    """Test all the base functionality of a Timestep

    All Timesteps must pass these tests!

    These test the Timestep independent of the Reader which it
    comes into contact with.  Failures here are the Timesteps fault.
    """
    # define the class made in test
    Timestep = mda.coordinates.base.Timestep
    name = "base"  # for error messages only
    size = 10  # size of arrays, 10 is enough to allow slicing etc
    # each coord is unique
    refpos = np.arange(size * 3, dtype=np.float32).reshape(size, 3) * 1.234
    refvel = np.arange(size * 3, dtype=np.float32).reshape(size, 3) * 2.345
    reffor = np.arange(size * 3, dtype=np.float32).reshape(size, 3) * 3.456

    # If you can set box, what the underlying unitcell should be
    # if dimensions are:
    newbox = np.array([10., 11., 12., 90., 90., 90.])
    unitcell = np.array([10., 11., 12., 90., 90., 90.])
    ref_volume = 1320.  # what the volume is after setting newbox
    uni_args = None

    @pytest.fixture()
    def ts(self):
        ts = self.Timestep(self.size)
        ts.frame += 1
        ts.positions = self.refpos
        ts.dimensions = self.newbox

        return ts

    def test_getitem(self, ts):
        assert_equal(ts[1], self.refpos[1])

    def test_getitem_neg(self, ts):
        assert_equal(ts[-1], self.refpos[-1])

    def test_getitem_neg_IE(self, ts):
        with pytest.raises(IndexError):
            ts.__getitem__(-(self.size + 1))


    def test_getitem_pos_IE(self, ts):
        with pytest.raises(IndexError):
            ts.__getitem__((self.size + 1))

    def test_getitem_slice(self, ts):
        assert_equal(len(ts[:2]), len(self.refpos[:2]))
        assert_allclose(ts[:2], self.refpos[:2])

    def test_getitem_slice2(self, ts):
        assert_equal(len(ts[1::2]), len(self.refpos[1::2]))
        assert_allclose(ts[1::2], self.refpos[1::2])

    def test_getitem_ndarray(self, ts):
        sel = np.array([0, 1, 4])
        assert_equal(len(ts[sel]), len(self.refpos[sel]))
        assert_allclose(ts[sel], self.refpos[sel])

    def test_getitem_TE(self, ts):
        with pytest.raises(TypeError):
            ts.__getitem__('string')

    def test_len(self, ts):
        assert_equal(len(ts), self.size)

    def test_iter(self, ts):
        for a, b in zip(ts, self.refpos):
            assert_allclose(a, b)
        assert_equal(len(list(ts)), self.size)

    def test_repr(self, ts):
        assert_equal(type(repr(ts)), str)

    # Dimensions has 2 possible cases
    # Timestep doesn't do dimensions,
    # should raise NotImplementedError for .dimension and .volume
    # Timestep does do them, should return values properly
    def test_dimensions(self, ts):
        assert_allclose(ts.dimensions, self.newbox)

    @pytest.mark.parametrize('dtype', (int, np.float32, np.float64))
    def test_dimensions_set_box(self, ts, dtype):
        ts.dimensions = self.newbox.astype(dtype)
        assert ts.dimensions.dtype == np.float32
        assert_allclose(ts.dimensions, self.newbox)

    def test_volume(self, ts):
        ts.dimensions = self.newbox
        assert_equal(ts.volume, self.ref_volume)

    def test_triclinic_vectors(self, ts):
        assert_allclose(ts.triclinic_dimensions,
                        triclinic_vectors(ts.dimensions))

    def test_set_triclinic_vectors(self, ts):
        ref_vec = triclinic_vectors(self.newbox)
        ts.triclinic_dimensions = ref_vec
        assert_equal(ts.dimensions, self.newbox)

    def test_coordinate_getter_shortcuts(self, ts):
        """testing that reading _x, _y, and _z works as expected
        # (Issue 224) (TestTimestep)"""
        assert_equal(ts._x, ts._pos[:, 0])
        assert_equal(ts._y, ts._pos[:, 1])
        assert_equal(ts._z, ts._pos[:, 2])

    def test_coordinate_setter_shortcuts(self, ts):
        # Check that _x _y and _z are read only
        for coordinate in ('_x', '_y', '_z'):
            random_positions = np.arange(self.size).astype(np.float32)
            with pytest.raises(AttributeError):
                setattr(ts, coordinate, random_positions)

    # n_atoms should be a read only property
    # all Timesteps require this attribute
    def test_n_atoms(self, ts):
        assert_equal(ts.n_atoms, ts._n_atoms)

    def test_n_atoms_readonly(self, ts):
        with pytest.raises(AttributeError):
            ts.__setattr__('n_atoms', 20)

    def test_n_atoms_presence(self, ts):
        assert_equal(hasattr(ts, '_n_atoms'), True)

    def test_unitcell_presence(self, ts):
        assert_equal(hasattr(ts, '_unitcell'), True)

    def test_data_presence(self, ts):
        assert_equal(hasattr(ts, 'data'), True)
        assert_equal(isinstance(ts.data, dict), True)

    def test_allocate_velocities(self, ts):
        assert_equal(ts.has_velocities, False)
        with pytest.raises(mda.NoDataError):
            getattr(ts, 'velocities')

        ts.has_velocities = True
        assert_equal(ts.has_velocities, True)
        assert_equal(ts.velocities.shape, (ts.n_atoms, 3))

    def test_allocate_forces(self, ts):
        assert_equal(ts.has_forces, False)
        with pytest.raises(mda.NoDataError):
            getattr(ts, 'forces')

        ts.has_forces = True
        assert_equal(ts.has_forces, True)
        assert_equal(ts.forces.shape, (ts.n_atoms, 3))

    def test_velocities_remove(self):
        ts = self.Timestep(10, velocities=True)
        ts.frame += 1
        assert_equal(ts.has_velocities, True)

        ts.has_velocities = False
        assert_equal(ts.has_velocities, False)
        with pytest.raises(mda.NoDataError):
            getattr(ts, 'velocities')

    def test_forces_remove(self):
        ts = self.Timestep(10, forces=True)
        ts.frame += 1
        assert_equal(ts.has_forces, True)

        ts.has_forces = False
        assert_equal(ts.has_forces, False)
        with pytest.raises(mda.NoDataError):
            getattr(ts, 'forces')

    def test_check_ts(self):
        with pytest.raises(ValueError):
            self.Timestep.from_coordinates(None, None, None)

    def _from_coords(self, p, v, f):
        posi = self.refpos if p else None
        velo = self.refvel if v else None
        forc = self.reffor if f else None

        ts = self.Timestep.from_coordinates(posi, velo, forc)

        return ts

    @pytest.mark.parametrize('p, v, f', filter(any,
                                               itertools.product([True, False],
                                                                 repeat=3)))
    def test_from_coordinates(self, p, v, f):
        ts = self._from_coords(p, v, f)

        if p:
            assert_array_almost_equal(ts.positions, self.refpos)
        else:
            with pytest.raises(mda.NoDataError):
                getattr(ts, 'positions')
        if v:
            assert_array_almost_equal(ts.velocities, self.refvel)
        else:
            with pytest.raises(mda.NoDataError):
                getattr(ts, 'velocities')
        if f:
            assert_array_almost_equal(ts.forces, self.reffor)
        else:
            with pytest.raises(mda.NoDataError):
                getattr(ts, 'forces')

    def test_from_coordinates_mismatch(self):
        velo = self.refvel[:2]

        with pytest.raises(ValueError):
            self.Timestep.from_coordinates(self.refpos, velo)

    def test_from_coordinates_nodata(self):
        with pytest.raises(ValueError):
            self.Timestep.from_coordinates()

    # Time related tests
    def test_supply_dt(self):
        # Check that this gets stored in data properly
        ts = self.Timestep(20, dt=0.04)

        assert_equal(ts.data['dt'], 0.04)
        assert_equal(ts.dt, 0.04)

    def test_redefine_dt(self):
        ts = self.Timestep(20, dt=0.04)
        assert_equal(ts.data['dt'], 0.04)
        assert_equal(ts.dt, 0.04)
        ts.dt = refdt = 0.46
        assert_equal(ts.data['dt'], refdt)
        assert_equal(ts.dt, refdt)

    def test_delete_dt(self):
        ts = self.Timestep(20, dt=0.04)
        assert_equal(ts.data['dt'], 0.04)
        assert_equal(ts.dt, 0.04)
        del ts.dt
        assert_equal('dt' in ts.data, False)
        assert_equal(ts.dt, 1.0)  # default value

    def test_supply_time_offset(self):
        ts = self.Timestep(20, time_offset=100.0)

        assert_equal(ts.data['time_offset'], 100.0)

    def test_time(self):
        ts = self.Timestep(20)
        ts.frame = 0
        ts.time = reftime = 1234.0

        assert_equal(ts.time, reftime)

    def test_delete_time(self):
        ts = self.Timestep(20)
        ts.frame = 0
        ts.time = reftime = 1234.0

        assert_equal(ts.time, reftime)
        del ts.time
        assert_equal(ts.time, 0.0)  # default to 1.0 (dt) * 0 (frame)

    def test_time_with_offset(self):
        ref_offset = 2345.0
        ts = self.Timestep(20, time_offset=ref_offset)
        ts.frame = 0
        ts.time = reftime = 1234.0

        assert_equal(ts.time, reftime + ref_offset)

    def test_dt(self):
        ref_dt = 45.0
        ts = self.Timestep(20, dt=ref_dt)

        for i in range(10):
            ts.frame = i
            assert_equal(ts.time, i * ref_dt)

    def test_change_dt(self):
        ref_dt = 45.0
        ts = self.Timestep(20, dt=ref_dt)

        for i in range(10):
            ts.frame = i
            assert_equal(ts.time, i * ref_dt)

        ts.dt = ref_dt = 77.0

        for i in range(10):
            ts.frame = i
            assert_equal(ts.time, i * ref_dt)

    def test_dt_with_offset(self):
        ref_dt = 45.0
        ref_offset = 2345.0
        ts = self.Timestep(20, dt=ref_dt, time_offset=ref_offset)

        for i in range(10):
            ts.frame = i
            assert_equal(ts.time, i * ref_dt + ref_offset)

    def test_time_overrides_dt_with_offset(self):
        ref_dt = 45.0
        ref_offset = 2345.0
        ts = self.Timestep(20, dt=ref_dt, time_offset=ref_offset)

        ts.frame = 0
        ts.time = reftime = 456.7

        assert_equal(ts.time, reftime + ref_offset)

    def _check_copy(self, name, ref_ts):
        """Check basic copy"""
        ts2 = ref_ts.copy()

        err_msg = ("Timestep copy failed for format {form}"
                   " on attribute {att}")

        # eq method checks:
        # - frame
        # - n_atoms
        # - positions, vels and forces
        assert ref_ts == ts2

        if not ref_ts.dimensions is None:
            assert_array_almost_equal(ref_ts.dimensions, ts2.dimensions,
                                      decimal=4)
        else:
            assert ref_ts.dimensions == ts2.dimensions

        # Check things not covered by eq
        for d in ref_ts.data:
            assert d in ts2.data
            if isinstance(ref_ts.data[d], np.ndarray):
                assert_array_almost_equal(
                    ref_ts.data[d], ts2.data[d])
            else:
                assert ref_ts.data[d] == ts2.data[d]

    def _check_independent(self, name, ts):
        """Check that copies made are independent"""
        ts2 = ts.copy()

        if ts.has_positions:
            self._check_array(ts.positions, ts2.positions)
        if ts.has_velocities:
            self._check_array(ts.velocities, ts2.velocities)
        if ts.has_forces:
            self._check_array(ts.forces, ts2.forces)
        if not ts.dimensions is None:
            self._check_array(ts.dimensions, ts2.dimensions)

    def _check_array(self, arr1, arr2):
        """Check modifying one array doesn't change other"""
        ref = arr1.copy()
        arr2 += 1.0
        assert_array_almost_equal(ref, arr1)

    def _check_copy_slice_indices(self, name, ts):
        sl = slice(0, len(ts), 3)
        ts2 = ts.copy_slice(sl)
        self._check_slice(ts, ts2, sl)

    def _check_copy_slice_slice(self, name, ts):
        sl = [0, 1, 3, 5, 6, 7]
        ts2 = ts.copy_slice(sl)
        self._check_slice(ts, ts2, sl)

    def _check_npint_slice(self, name, ts):
        for integers in [np.byte, np.short, np.intc, np.int_, np.longlong,
                         np.intp, np.int8, np.int16, np.int32, np.int64,
                         np.ubyte, np.ushort, np.uintc, np.ulonglong,
                         np.uintp, np.uint8, np.uint16, np.uint32, np.uint64]:
            sl = slice(1, 2, 1)
            ts2 = ts.copy_slice(slice(integers(1), integers(2), integers(1)))
            self._check_slice(ts, ts2, sl)

    def _check_slice(self, ts1, ts2, sl):
        if ts1.has_positions:
            assert_array_almost_equal(ts1.positions[sl], ts2.positions)
        if ts1.has_velocities:
            assert_array_almost_equal(ts1.velocities[sl], ts2.velocities)
        if ts1.has_forces:
            assert_array_almost_equal(ts1.forces[sl], ts2.forces)

    @pytest.mark.parametrize('func', [
        _check_copy,
        _check_independent,
        _check_copy_slice_indices,
        _check_copy_slice_slice,
        _check_npint_slice
    ])
    def test_copy(self, func, ts):
        if self.uni_args is None:
            return
        u = mda.Universe(*self.uni_args)  # pylint: disable=not-an-iterable
        ts = u.trajectory.ts
        func(self, self.name, ts)

    @pytest.fixture(params=filter(any,
                                  itertools.product([True, False], repeat=3)))
    def some_ts(self, request):
        p, v, f = request.param
        return self._from_coords(p, v, f)

    @pytest.mark.parametrize('func', [
        _check_copy,
        _check_independent,
        _check_copy_slice_indices,
        _check_copy_slice_slice,
        _check_npint_slice
    ])
    def test_copy_slice(self, func, some_ts):
        func(self, self.name, some_ts)

    def test_bad_slice(self, some_ts):
        sl = ['this', 'is', 'silly']
        with pytest.raises(TypeError):
            some_ts.copy_slice(sl)

    def test_from_timestep(self, some_ts):
        ts = some_ts
        ts2 = self.Timestep.from_timestep(ts)

        assert_timestep_almost_equal(ts, ts2)

    def _get_pos(self):
        # Get generic reference positions
        return np.arange(30).reshape(10, 3) * 1.234

    @pytest.mark.parametrize('p, v, f', filter(any,
                                               itertools.product([True, False],
                                                                 repeat=3)))
    def test_check_equal(self, p, v, f):
        ts1 = self.Timestep(self.size,
                            positions=p,
                            velocities=v,
                            forces=f)
        ts2 = self.Timestep(self.size,
                            positions=p,
                            velocities=v,
                            forces=f)
        if p:
            ts1.positions = self.refpos.copy()
            ts2.positions = self.refpos.copy()
        if v:
            ts1.velocities = self.refvel.copy()
            ts2.velocities = self.refvel.copy()
        if f:
            ts1.forces = self.reffor.copy()
            ts2.forces = self.reffor.copy()

        assert_timestep_equal(ts1, ts2)

    def test_wrong_class_equality(self):
        ts1 = self.Timestep(self.size)
        ts1.positions = self._get_pos()

        b = tuple([0, 1, 2, 3])

        assert ts1 != b
        assert b != ts1

    def test_wrong_frame_equality(self):
        ts1 = self.Timestep(self.size)
        ts1.positions = self._get_pos()

        ts2 = self.Timestep(self.size)
        ts2.positions = self._get_pos()
        ts2.frame = 987

        assert ts1 != ts2
        assert ts2 != ts1

    def test_wrong_n_atoms_equality(self):
        ts1 = self.Timestep(self.size)
        ts1.positions = self._get_pos()

        ts3 = self.Timestep(self.size * 2)

        assert ts1 != ts3
        assert ts3 != ts1

    def test_wrong_pos_equality(self):
        ts1 = self.Timestep(self.size)
        ts1.positions = self._get_pos()

        ts2 = self.Timestep(self.size)
        ts2.positions = self._get_pos() + 1.0

        assert ts1 != ts2
        assert ts2 != ts1

    def test_no_pos_inequality(self):
        ts1 = self.Timestep(self.size, positions=False)
        ts2 = self.Timestep(self.size)

        assert ts1 != ts2
        assert ts2 != ts1

    def test_check_vels_equality(self):
        ts1 = self.Timestep(self.size, velocities=True)
        ts2 = self.Timestep(self.size, velocities=True)

        ts1.velocities = self._get_pos()
        ts2.velocities = self._get_pos()

        assert ts1 == ts2
        assert ts2 == ts1

    def test_check_mismatched_vels_equality(self):
        ts1 = self.Timestep(self.size, velocities=True)
        ts2 = self.Timestep(self.size, velocities=False)

        ts1.velocities = self._get_pos()

        assert ts1 != ts2
        assert ts2 != ts1

    def test_check_wrong_vels_equality(self):
        ts1 = self.Timestep(self.size, velocities=True)
        ts2 = self.Timestep(self.size, velocities=True)

        ts1.velocities = self._get_pos()
        ts2.velocities = self._get_pos() + 1.0

        assert ts1 != ts2
        assert ts2 != ts1

    def test_check_forces_equality(self):
        ts1 = self.Timestep(self.size, forces=True)
        ts2 = self.Timestep(self.size, forces=True)

        ts1.forces = self._get_pos()
        ts2.forces = self._get_pos()

        assert ts1 == ts2
        assert ts2 == ts1

    def test_check_mismatched_forces_equality(self):
        ts1 = self.Timestep(self.size, forces=True)
        ts2 = self.Timestep(self.size, forces=False)

        ts1.forces = self._get_pos()

        assert ts1 != ts2
        assert ts2 != ts1

    def test_check_wrong_forces_equality(self):
        ts1 = self.Timestep(self.size, forces=True)
        ts2 = self.Timestep(self.size, forces=True)

        ts1.forces = self._get_pos()
        ts2.forces = self._get_pos() + 1.0

        assert ts1 != ts2
        assert ts2 != ts1

    @pytest.mark.parametrize('dim1', [None, [2., 2., 2., 90., 90., 90.]])
    def test_dims_mismatch_inequality(self, dim1):
        ts1 = self.Timestep(self.size)
        ts1.dimensions = dim1
        ts2 = self.Timestep(self.size)
        ts2.dimensions = [1., 1., 1., 90., 90., 90.]

        assert ts1 != ts2
        assert ts2 != ts1


# TODO: Merge this into generic Reader tests
# These tests are all included in BaseReaderTest
# Once Readers use that TestClass, delete this one

class TestBaseTimestepInterface(object):
    """Test the Timesteps created by Readers

    This checks that Readers are creating correct Timestep objects,
    ensuring a consistent interface when using Timestep objects.

    Failures here are the Reader's fault

    See Issue #250 for discussion
    """
    @pytest.fixture(params=(
        (XYZ_five, INPCRD, None, None),
        (PSF, DCD, None, None),
        (DLP_CONFIG, None, 'CONFIG', None),
        (DLP_HISTORY, None, 'HISTORY', None),
        (DMS, None, None, None),
        (GRO, None, None, None),
        (XYZ_five, INPCRD, None, None),
        (LAMMPSdata, None, None, None),
        (mol2_molecules, None, None, None),
        (PDB_small, None, None, None),
        (PQR, None, None, None),
        (PDBQT_input, None, None, None),
        (PRM, TRJ, None, None),
        (GRO, XTC, None, None),
        (TRZ_psf, TRZ, None, None),
        (GRO, TRR, None, None),
        (GMS_ASYMOPT, GMS_ASYMOPT, 'GMS', 'GMS'),
        (LAMMPSdata2, LAMMPSdcd2, 'LAMMPS', 'DATA'),
        (PRMncdf, NCDF, None, None),
    ))
    def universe(self, request):
        topology, trajectory, trajectory_format, topology_format = request.param
        if trajectory_format is not None and topology_format is not None:
            return mda.Universe(topology, trajectory, format=trajectory_format,
                                topology_format=topology_format)

        if trajectory is not None:
            return mda.Universe(topology, trajectory)
        else:
            return mda.Universe(topology, format=trajectory_format)

    def test_frame(self, universe):
        assert_equal(universe.trajectory.ts.frame, 0)

    def test_dt(self, universe):
        assert_equal(universe.trajectory.dt, universe.trajectory.ts.dt)


@pytest.mark.parametrize('uni', [
    [(PSF, DCD), {}],  # base Timestep
    [(DLP_CONFIG,), {'format': 'CONFIG'}],  # DLPoly
    [(DMS,), {}],  # DMS
    [(GRO,), {}],  # GRO
    [(np.zeros((1, 30, 3)),), {}],  # memory
    [(PRM, TRJ), {}],  # TRJ
    [(TRZ_psf, TRZ), {}],  # TRZ
])
def test_atomgroup_dims_access(uni):
    uni_args, uni_kwargs = uni
    # check that AtomGroup.dimensions always returns a copy
    u = mda.Universe(*uni_args, **uni_kwargs)

    ag = u.atoms[:10]

    dims = ag.dimensions
    ts = u.trajectory.ts

    # dimensions from AtomGroup should be equal
    assert_equal(dims, ts.dimensions)
    # but if np array shouldn't be the same object, i.e. it should have been copied
    if dims is not None:
        assert dims is not ts.dimensions


class TimeTimestepNoBox:
    @staticmethod
    @pytest.fixture
    def ts():
        ts = mda.coordinates.base.Timestep(20)

        ts.dimensions = None

        return ts

    def test_Noneness(self, ts):
        assert ts.dimensions is None

    def test_triclinic_Noneness(self, ts):
        assert ts.triclinic_dimensions is None

    def test_set_dimensions(self, ts):
        # None to something
        box = np.array([10, 10, 10, 90, 90, 90])
        ts.dimensions = box

        assert_array_equal(ts.dimensions, box)

    def test_set_triclinic_dimensions(self, ts):
        # None to something via triclinic
        box = np.array([[10, 0, 0], [5, 10, 0], [0, 5, 10]])

        ts.triclinic_dimensions = box

        assert_array_equal(ts.triclinic_dimensions, box)

    def test_set_None(self, ts):
        ts.dimensions = [10, 10, 10, 90, 90, 90]

        assert not ts.dimensions is None

        ts.dimensions = None

        assert ts.dimensions is None
        assert ts.triclinic_dimenions is None

    def test_set_triclinic_None(self, ts):
        ts.dimensions = [10, 10, 10, 90, 90, 90]

        assert not ts.dimensions is None

        ts.triclinic_dimensions = None

        assert ts.dimensions is None
        assert ts.triclinic_dimenions is None

    def test_set_zero_box(self, ts):
        ts.dimensions = np.zeros(6)

        assert ts.dimensions is None
        assert ts.triclinic_dimenions is None

    def test_set_zero_box_triclinic(self, ts):
        ts.triclinic_dimensions = np.zeros((3, 3))

        assert ts.dimensions is None
        assert ts.triclinic_dimenions is None



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import os

import MDAnalysis
import MDAnalysis as mda
from MDAnalysis.analysis import rms, align

from numpy.testing import assert_equal, assert_almost_equal

import numpy as np
import pytest

from MDAnalysis.exceptions import SelectionError, NoDataError
from MDAnalysisTests.datafiles import GRO, XTC, rmsfArray, PSF, DCD


class Testrmsd(object):
    shape = (5, 3)
    # vectors with length one
    ones = np.ones(shape) / np.sqrt(3)

    @pytest.fixture()
    def a(self):
        return self.ones * np.arange(1, 6)[:, np.newaxis]

    @pytest.fixture()
    def b(self, a):
        return a + self.ones

    @pytest.fixture()
    def u(self):
        u = mda.Universe(PSF, DCD)
        return u

    @pytest.fixture()
    def u2(self):
        u = mda.Universe(PSF, DCD)
        return u

    @pytest.fixture()
    def p_first(self, u):
        u.trajectory[0]
        return u.select_atoms('protein')

    @pytest.fixture()
    def p_last(self, u2):
        u2.trajectory[-1]
        return u2.select_atoms('protein')

    # internal test
    def test_p_frames(self, p_first, p_last):
        # check that these fixtures are really different
        assert p_first.universe.trajectory.ts.frame != p_last.universe.trajectory.ts.frame
        assert not np.allclose(p_first.positions, p_last.positions)

    def test_no_center(self, a, b):
        rmsd = rms.rmsd(a, b, center=False)
        assert_almost_equal(rmsd, 1.0)

    def test_center(self, a, b):
        rmsd = rms.rmsd(a, b, center=True)
        assert_almost_equal(rmsd, 0.0)

    def test_list(self, a, b):
        rmsd = rms.rmsd(a.tolist(),
                        b.tolist(),
                        center=False)
        assert_almost_equal(rmsd, 1.0)

    def test_superposition(self, a, b, u):
        bb = u.atoms.select_atoms('backbone')
        a = bb.positions.copy()
        u.trajectory[-1]
        b = bb.positions.copy()
        rmsd = rms.rmsd(a, b, superposition=True)
        assert_almost_equal(rmsd, 6.820321761927005)

    def test_weights(self, a, b):
        weights = np.zeros(len(a))
        weights[0] = 1
        weights[1] = 1
        weighted = rms.rmsd(a, b, weights=weights)
        firstCoords = rms.rmsd(a[:2], b[:2])
        assert_almost_equal(weighted, firstCoords)

    def test_weights_and_superposition_1(self, u):
        weights = np.ones(len(u.trajectory[0]))
        weighted = rms.rmsd(u.trajectory[0], u.trajectory[1],
                            weights=weights, superposition=True)
        firstCoords = rms.rmsd(u.trajectory[0], u.trajectory[1],
                               superposition=True)
        assert_almost_equal(weighted, firstCoords, decimal=5)

    def test_weights_and_superposition_2(self, u):
        weights = np.zeros(len(u.trajectory[0]))
        weights[:100] = 1
        weighted = rms.rmsd(u.trajectory[0], u.trajectory[-1],
                            weights=weights, superposition=True)
        firstCoords = rms.rmsd(u.trajectory[0][:100],
                               u.trajectory[-1][:100],
                               superposition=True)
        # very close to zero, change significant decimal places to 5
        assert_almost_equal(weighted, firstCoords, decimal=5)

    def test_unequal_shape(self):
        a = np.ones((4, 3))
        b = np.ones((5, 3))
        with pytest.raises(ValueError):
            rms.rmsd(a, b)

    def test_wrong_weights(self, a, b):
        w = np.ones(2)
        with pytest.raises(ValueError):
            rms.rmsd(a, b, w)

    def test_with_superposition_smaller(self, p_first, p_last):
        A = p_first.positions
        B = p_last.positions
        rmsd = rms.rmsd(A, B)
        rmsd_superposition = rms.rmsd(A, B, center=True, superposition=True)
        # by design the super positioned rmsd is smaller
        assert rmsd > rmsd_superposition

    def test_with_superposition_equal(self, p_first, p_last):
        align.alignto(p_first, p_last)
        A = p_first.positions
        B = p_last.positions
        rmsd = rms.rmsd(A, B)
        rmsd_superposition = rms.rmsd(A, B, center=True, superposition=True)
        assert_almost_equal(rmsd, rmsd_superposition, decimal=6)


class TestRMSD(object):
    @pytest.fixture()
    def universe(self):
        return MDAnalysis.Universe(PSF, DCD)

    @pytest.fixture()
    def outfile(self, tmpdir):
        return os.path.join(str(tmpdir), 'rmsd.txt')

    @pytest.fixture()
    def correct_values(self):
        return [[0, 1, 0], [49, 50, 4.68953]]

    @pytest.fixture()
    def correct_values_mass(self):
        return [[0, 1, 0], [49, 50, 4.74920]]

    @pytest.fixture()
    def correct_values_mass_add_ten(self):
        return [[0, 1, 0.0632], [49, 50, 4.7710]]

    @pytest.fixture()
    def correct_values_group(self):
        return [[0, 1, 0, 0, 0],
                [49, 50, 4.7857, 4.7048, 4.6924]]

    @pytest.fixture()
    def correct_values_backbone_group(self):
        return [[0, 1, 0, 0, 0],
                [49, 50, 4.6997, 1.9154, 2.7139]]

    def test_rmsd(self, universe, correct_values):
        RMSD = MDAnalysis.analysis.rms.RMSD(universe, select='name CA')
        RMSD.run(step=49)
        assert_almost_equal(RMSD.results.rmsd, correct_values, 4,
                            err_msg="error: rmsd profile should match" +
                            "test values")

    def test_rmsd_frames(self, universe, correct_values):
        RMSD = MDAnalysis.analysis.rms.RMSD(universe, select='name CA')
        RMSD.run(frames=[0, 49])
        assert_almost_equal(RMSD.results.rmsd, correct_values, 4,
                            err_msg="error: rmsd profile should match" +
                            "test values")

    def test_rmsd_unicode_selection(self, universe, correct_values):
        RMSD = MDAnalysis.analysis.rms.RMSD(universe, select=u'name CA')
        RMSD.run(step=49)
        assert_almost_equal(RMSD.results.rmsd, correct_values, 4,
                            err_msg="error: rmsd profile should match" +
                            "test values")

    def test_rmsd_atomgroup_selections(self, universe):
        # see Issue #1684
        R1 = MDAnalysis.analysis.rms.RMSD(universe.atoms,
                                          select="resid 1-30").run()
        R2 = MDAnalysis.analysis.rms.RMSD(universe.atoms.select_atoms("name CA"),
                                          select="resid 1-30").run()
        assert not np.allclose(R1.results.rmsd[:, 2], R2.results.rmsd[:, 2])

    def test_rmsd_single_frame(self, universe):
        RMSD = MDAnalysis.analysis.rms.RMSD(universe, select='name CA',
                                            ).run(start=5, stop=6)
        single_frame = [[5, 6, 0.91544906]]
        assert_almost_equal(RMSD.results.rmsd, single_frame, 4,
                            err_msg="error: rmsd profile should match" +
                            "test values")

    def test_mass_weighted(self, universe, correct_values):
        # mass weighting the CA should give the same answer as weighing
        # equally because all CA have the same mass
        RMSD = MDAnalysis.analysis.rms.RMSD(universe, select='name CA',
                                            weights='mass').run(step=49)

        assert_almost_equal(RMSD.results.rmsd, correct_values, 4,
                            err_msg="error: rmsd profile should match"
                            "test values")

    def test_custom_weighted(self, universe, correct_values_mass):
        RMSD = MDAnalysis.analysis.rms.RMSD(universe, weights="mass").run(step=49)

        assert_almost_equal(RMSD.results.rmsd, correct_values_mass, 4,
                            err_msg="error: rmsd profile should match"
                            "test values")

    def test_weights_mass_is_mass_weighted(self, universe):
        RMSD_mass = MDAnalysis.analysis.rms.RMSD(universe,
                                                 weights="mass").run(step=49)
        RMSD_cust = MDAnalysis.analysis.rms.RMSD(universe,
                                                 weights=universe.atoms.masses).run(step=49)
        assert_almost_equal(RMSD_mass.results.rmsd, RMSD_cust.results.rmsd, 4,
                            err_msg="error: rmsd profiles should match for 'mass' "
                            "and universe.atoms.masses")

    def test_custom_weighted_list(self, universe, correct_values_mass):
        weights = universe.atoms.masses
        RMSD = MDAnalysis.analysis.rms.RMSD(universe,
                                            weights=list(weights)).run(step=49)
        assert_almost_equal(RMSD.results.rmsd, correct_values_mass, 4,
                            err_msg="error: rmsd profile should match" +
                            "test values")

    def test_custom_groupselection_weights_applied_1D_array(self, universe):
        RMSD = MDAnalysis.analysis.rms.RMSD(universe,
                                            select='backbone',
                                            groupselections=['name CA and resid 1-5', 'name CA and resid 1'],
                                            weights=None,
                                            weights_groupselections=[[1, 0, 0, 0, 0], None]).run(step=49)

        assert_almost_equal(RMSD.results.rmsd.T[3], RMSD.results.rmsd.T[4], 4,
                            err_msg="error: rmsd profile should match "
                            "for applied weight array and selected resid")

    def test_custom_groupselection_weights_applied_mass(self, universe, correct_values_mass):
        RMSD = MDAnalysis.analysis.rms.RMSD(universe,
                                            select='backbone',
                                            groupselections=['all', 'all'],
                                            weights=None,
                                            weights_groupselections=['mass',
                                                                     universe.atoms.masses]).run(step=49)

        assert_almost_equal(RMSD.results.rmsd.T[3], RMSD.results.rmsd.T[4], 4,
                            err_msg="error: rmsd profile should match "
                            "between applied mass and universe.atoms.masses")

    def test_rmsd_scalar_weights_raises_ValueError(self, universe):
        with pytest.raises(ValueError):
            RMSD = MDAnalysis.analysis.rms.RMSD(
                universe, weights=42)

    def test_rmsd_string_weights_raises_ValueError(self, universe):
        with pytest.raises(ValueError):
            RMSD = MDAnalysis.analysis.rms.RMSD(
                universe, weights="Jabberwock")

    def test_rmsd_mismatched_weights_raises_ValueError(self, universe):
        with pytest.raises(ValueError):
            RMSD = MDAnalysis.analysis.rms.RMSD(
                universe, weights=universe.atoms.masses[:-1])

    def test_rmsd_misuse_weights_for_groupselection_raises_TypeError(self, universe):
        with pytest.raises(TypeError):
            RMSD = MDAnalysis.analysis.rms.RMSD(
                universe, groupselections=['all'],
                weights=[universe.atoms.masses, universe.atoms.masses[:-1]])

    def test_rmsd_mismatched_weights_in_groupselection_raises_ValueError(self, universe):
        with pytest.raises(ValueError):
            RMSD = MDAnalysis.analysis.rms.RMSD(
                universe, groupselections=['all'],
                weights=universe.atoms.masses,
                weights_groupselections = [universe.atoms.masses[:-1]])

    def test_rmsd_list_of_weights_wrong_length(self, universe):
        with pytest.raises(ValueError):
            RMSD = MDAnalysis.analysis.rms.RMSD(
                universe, groupselections=['backbone', 'name CA'],
                weights='mass',
                weights_groupselections=[None])

    def test_rmsd_group_selections(self, universe, correct_values_group):
        RMSD = MDAnalysis.analysis.rms.RMSD(universe,
                                            groupselections=['backbone', 'name CA']
                                            ).run(step=49)
        assert_almost_equal(RMSD.results.rmsd, correct_values_group, 4,
                            err_msg="error: rmsd profile should match"
                            "test values")

    def test_rmsd_backbone_and_group_selection(self, universe,
                                               correct_values_backbone_group):
        RMSD = MDAnalysis.analysis.rms.RMSD(
            universe,
            reference=universe,
            select="backbone",
            groupselections=['backbone and resid 1:10',
                             'backbone and resid 10:20']).run(step=49)
        assert_almost_equal(
            RMSD.results.rmsd, correct_values_backbone_group, 4,
            err_msg="error: rmsd profile should match test values")

    def test_ref_length_unequal_len(self, universe):
        reference = MDAnalysis.Universe(PSF, DCD)
        reference.atoms = reference.atoms[:-1]
        with pytest.raises(SelectionError):
            RMSD = MDAnalysis.analysis.rms.RMSD(universe,
                                                reference=reference)

    def test_mass_mismatches(self, universe):
        reference = MDAnalysis.Universe(PSF, DCD)
        reference.atoms.masses = 10
        with pytest.raises(SelectionError):
            RMSD = MDAnalysis.analysis.rms.RMSD(universe,
                                                reference=reference)

    def test_ref_mobile_mass_mismapped(self, universe,correct_values_mass_add_ten):
        reference = MDAnalysis.Universe(PSF, DCD)
        universe.atoms.masses = universe.atoms.masses + 10
        RMSD = MDAnalysis.analysis.rms.RMSD(universe,
                                                reference=reference,
                                                select='all',
                                                weights='mass',
                                                tol_mass=100)
        RMSD.run(step=49)
        assert_almost_equal(RMSD.results.rmsd, correct_values_mass_add_ten, 4,
                            err_msg="error: rmsd profile should match "
                            "between true values and calculated values")

    def test_group_selections_unequal_len(self, universe):
        reference = MDAnalysis.Universe(PSF, DCD)
        reference.atoms[0].residue.resname = 'NOTMET'
        with pytest.raises(SelectionError):
            RMSD = MDAnalysis.analysis.rms.RMSD(universe,
                                                reference=reference,
                                                groupselections=['resname MET', 'type NH3'])

    def test_rmsd_attr_warning(self, universe):
        RMSD = MDAnalysis.analysis.rms.RMSD(
                universe, select='name CA').run(stop=2)

        wmsg = "The `rmsd` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(RMSD.rmsd, RMSD.results.rmsd)


class TestRMSF(object):
    @pytest.fixture()
    def universe(self):
        return mda.Universe(GRO, XTC)

    def test_rmsf(self, universe):
        rmsfs = rms.RMSF(universe.select_atoms('name CA'))
        rmsfs.run()
        test_rmsfs = np.load(rmsfArray)

        assert_almost_equal(rmsfs.results.rmsf, test_rmsfs, 5,
                            err_msg="error: rmsf profile should match test "
                            "values")

    def test_rmsf_single_frame(self, universe):
        rmsfs = rms.RMSF(universe.select_atoms('name CA')).run(start=5, stop=6)

        assert_almost_equal(rmsfs.results.rmsf, 0, 5,
                            err_msg="error: rmsfs should all be zero")

    def test_rmsf_identical_frames(self, universe, tmpdir):

        outfile = os.path.join(str(tmpdir), 'rmsf.xtc')

        # write a dummy trajectory of all the same frame
        with mda.Writer(outfile, universe.atoms.n_atoms) as W:
            for _ in range(universe.trajectory.n_frames):
                W.write(universe)

        universe = mda.Universe(GRO, outfile)
        rmsfs = rms.RMSF(universe.select_atoms('name CA'))
        rmsfs.run()
        assert_almost_equal(rmsfs.results.rmsf, 0, 5,
                            err_msg="error: rmsfs should all be 0")

    def test_rmsf_attr_warning(self, universe):
        rmsfs = rms.RMSF(universe.select_atoms('name CA')).run(stop=2)

        wmsg = "The `rmsf` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(rmsfs.rmsf, rmsfs.results.rmsf)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from numpy.testing import (
    assert_equal,
)
import pytest

import MDAnalysis as mda

from MDAnalysisTests.datafiles import PSF, DCD


# Legacy tests from before 363
@pytest.fixture()
def res():
    universe = mda.Universe(PSF, DCD)
    return universe.residues[100]


def test_type(res):
    assert isinstance(res, mda.core.groups.Residue)
    assert_equal(res.resname, "ILE")
    assert_equal(res.resid, 101)


def test_index(res):
    atom = res.atoms[2]
    assert isinstance(atom, mda.core.groups.Atom)
    assert_equal(atom.name, "CA")
    assert_equal(atom.index, 1522)
    assert_equal(atom.resid, 101)


def test_atom_order(res):
    assert_equal(res.atoms.indices,
                 sorted(res.atoms.indices))



import pytest
from numpy.testing import assert_almost_equal, assert_equal, assert_allclose
import numpy as np
import sys
import os
import MDAnalysis as mda
from MDAnalysis.coordinates.H5MD import HAS_H5PY
if HAS_H5PY:
    import h5py
from MDAnalysis.exceptions import NoDataError
from MDAnalysisTests import make_Universe
from MDAnalysisTests.datafiles import (H5MD_xvf, TPR_xvf, TRR_xvf,
                                       COORDINATES_TOPOLOGY,
                                       COORDINATES_H5MD)
from MDAnalysisTests.coordinates.base import (MultiframeReaderTest,
                                              BaseReference, BaseWriterTest,
                                              assert_timestep_almost_equal)


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class H5MDReference(BaseReference):
    """Reference synthetic trajectory that was
    copied from test_xdr.TRRReference"""

    def __init__(self):
        super(H5MDReference, self).__init__()
        self.trajectory = COORDINATES_H5MD
        self.topology = COORDINATES_TOPOLOGY
        self.reader = mda.coordinates.H5MD.H5MDReader
        self.writer = mda.coordinates.H5MD.H5MDWriter
        self.ext = 'h5md'
        self.prec = 3
        self.changing_dimensions = True

        self.first_frame.velocities = self.first_frame.positions / 10
        self.first_frame.forces = self.first_frame.positions / 100

        self.second_frame.velocities = self.second_frame.positions / 10
        self.second_frame.forces = self.second_frame.positions / 100

        self.last_frame.velocities = self.last_frame.positions / 10
        self.last_frame.forces = self.last_frame.positions / 100

        self.jump_to_frame.velocities = self.jump_to_frame.positions / 10
        self.jump_to_frame.forces = self.jump_to_frame.positions / 100

    def iter_ts(self, i):
        ts = self.first_frame.copy()
        ts.positions = 2**i * self.first_frame.positions
        ts.velocities = ts.positions / 10
        ts.forces = ts.positions / 100
        ts.time = i
        ts.frame = i
        return ts


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class TestH5MDReaderBaseAPI(MultiframeReaderTest):
    """Tests H5MDReader with with synthetic trajectory."""
    @staticmethod
    @pytest.fixture()
    def ref():
        return H5MDReference()

    def test_get_writer_1(self, ref, reader, tmpdir):
        with tmpdir.as_cwd():
            outfile = 'test-writer.' + ref.ext
            with reader.Writer(outfile) as W:
                assert_equal(isinstance(W, ref.writer), True)
                assert_equal(W.n_atoms, reader.n_atoms)

    def test_get_writer_2(self, ref, reader, tmpdir):
        with tmpdir.as_cwd():
            outfile = 'test-writer.' + ref.ext
            with reader.Writer(outfile, n_atoms=100) as W:
                assert_equal(isinstance(W, ref.writer), True)
                assert_equal(W.n_atoms, 100)

    def test_copying(self, ref, reader):
        # Issue #3664 - test not done in test_copying due to dependencies
        original = mda.coordinates.H5MD.H5MDReader(
                ref.trajectory, convert_units=False, dt=2,
                time_offset=10, foo="bar")
        copy = original.copy()

        assert original.format not in ('MEMORY', 'CHAIN')
        assert original.convert_units is False
        assert copy.convert_units is False
        assert original._ts_kwargs['time_offset'] == 10
        assert copy._ts_kwargs['time_offset'] == 10
        assert original._ts_kwargs['dt'] == 2
        assert copy._ts_kwargs['dt'] == 2

        assert original.ts.data['time_offset'] == 10
        assert copy.ts.data['time_offset'] == 10

        assert original.ts.data['dt'] == 2
        assert copy.ts.data['dt'] == 2

        assert copy._kwargs['foo'] == 'bar'

        # check coordinates
        assert original.ts.frame == copy.ts.frame
        assert_allclose(original.ts.positions, copy.ts.positions)

        original.next()
        copy.next()

        assert original.ts.frame == copy.ts.frame
        assert_allclose(original.ts.positions, copy.ts.positions)


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class TestH5MDWriterBaseAPI(BaseWriterTest):
    """Tests H5MDWriter base API with synthetic trajectory"""
    @staticmethod
    @pytest.fixture()
    def ref():
        return H5MDReference()

    def test_write_trajectory_atomgroup(self, ref, reader, universe, tmpdir):
        outfile = 'write-atoms-test.' + ref.ext
        with tmpdir.as_cwd():
            with ref.writer(outfile, universe.atoms.n_atoms,
                            velocities=True, forces=True) as w:
                for ts in universe.trajectory:
                    w.write(universe.atoms)
            self._check_copy(outfile, ref, reader)

    def test_write_trajectory_universe(self, ref, reader, universe, tmpdir):
        outfile = 'write-uni-test.' + ref.ext
        with tmpdir.as_cwd():
            with ref.writer(outfile, universe.atoms.n_atoms,
                            velocities=True, forces=True) as w:
                for ts in universe.trajectory:
                    w.write(universe)
            self._check_copy(outfile, ref, reader)


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class TestH5MDReaderWithRealTrajectory(object):

    prec = 3
    ext = 'h5md'

    @pytest.fixture(scope='class')
    def universe(self):
        return mda.Universe(TPR_xvf, H5MD_xvf)

    @pytest.fixture()
    def h5md_file(self):
        return h5py.File(H5MD_xvf, 'r')

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir.join('h5md-reader-test.' + self.ext))

    def test_n_frames(self, universe):
        assert len(universe.trajectory) == 3

    def test_positions(self, universe):
        universe.trajectory[0]
        assert_almost_equal(universe.atoms.positions[0],
                            [32.309906, 13.77798, 14.372463],
                            decimal=self.prec)
        assert_almost_equal(universe.atoms.positions[42],
                            [28.116928, 19.405945, 19.647358],
                            decimal=self.prec)
        assert_almost_equal(universe.atoms.positions[10000],
                            [44.117805, 50.442093, 23.299038],
                            decimal=self.prec)

        universe.trajectory[1]
        assert_almost_equal(universe.atoms.positions[0],
                            [30.891968, 13.678971, 13.6000595],
                            decimal=self.prec)
        assert_almost_equal(universe.atoms.positions[42],
                            [27.163246, 19.846561, 19.3582],
                            decimal=self.prec)
        assert_almost_equal(universe.atoms.positions[10000],
                            [45.869278, 5.0342298, 25.460655],
                            decimal=self.prec)
        universe.trajectory[2]
        assert_almost_equal(universe.atoms.positions[0],
                            [31.276512, 13.89617, 15.015897],
                            decimal=self.prec)
        assert_almost_equal(universe.atoms.positions[42],
                            [28.567991, 20.56532, 19.40814],
                            decimal=self.prec)
        assert_almost_equal(universe.atoms.positions[10000],
                            [39.713223,  6.127234, 18.284992],
                            decimal=self.prec)

    def test_h5md_velocities(self, universe):
        universe.trajectory[0]
        assert_almost_equal(universe.atoms.velocities[0],
                            [-2.697732, 0.613568, 0.14334752],
                            decimal=self.prec)
        universe.trajectory[1]
        assert_almost_equal(universe.atoms.velocities[42],
                            [-6.8698354, 7.834235, -8.114698],
                            decimal=self.prec)
        universe.trajectory[2]
        assert_almost_equal(universe.atoms.velocities[10000],
                            [9.799492, 5.631466, 6.852126],
                            decimal=self.prec)

    def test_h5md_forces(self, universe):
        universe.trajectory[0]
        assert_almost_equal(universe.atoms.forces[0],
                            [20.071287, -155.2285, -96.72112],
                            decimal=self.prec)
        universe.trajectory[1]
        assert_almost_equal(universe.atoms.forces[42],
                            [-4.1959066, -31.31548, 22.663044],
                            decimal=self.prec)
        universe.trajectory[2]
        assert_almost_equal(universe.atoms.forces[10000],
                            [-41.43743, 83.35207, 62.94751],
                            decimal=self.prec)

    def test_h5md_dimensions(self, universe):
        universe.trajectory[0]
        assert_almost_equal(universe.trajectory.ts.dimensions,
                            [52.763, 52.763, 52.763, 90., 90., 90.],
                            decimal=self.prec)
        universe.trajectory[1]
        assert_almost_equal(universe.trajectory.ts.dimensions,
                            [52.807877, 52.807877, 52.807877, 90., 90., 90.],
                            decimal=self.prec)
        universe.trajectory[2]
        assert_almost_equal(universe.trajectory.ts.dimensions,
                            [52.839806, 52.839806, 52.839806, 90., 90., 90.],
                            decimal=self.prec)

    def test_h5md_data_step(self, universe):
        for ts, step in zip(universe.trajectory, (0, 25000, 50000)):
            assert_equal(ts.data['step'], step)

    def test_rewind(self, universe):
        universe.trajectory[1]
        universe.trajectory.rewind()
        assert universe.trajectory.ts.frame == 0

    def test_next(self, universe):
        universe.trajectory.rewind()
        universe.trajectory.next()
        assert universe.trajectory.ts.frame == 1

    def test_jump_last_frame(self, universe):
        universe.trajectory[-1]
        assert universe.trajectory.ts.frame == 2

    @pytest.mark.parametrize("start, stop, step", ((0, 2, 1),
                                                   (1, 2, 1)))
    def test_slice(universe, start, stop, step):
        frames = [universe.trajectory.ts.frame
                  for ts in universe.trajectory[start:stop:step]]
        assert_equal(frames, np.arange(start, stop, step))

    @pytest.mark.parametrize("start, stop, step", ((0, 2, 1),
                                                   (1, 2, 1)))
    def test_slice(self, universe, start, stop, step):
        frames = [universe.trajectory.ts.frame
                  for ts in universe.trajectory[start:stop:step]]
        assert_equal(frames, np.arange(start, stop, step))

    @pytest.mark.parametrize("array_like", [list, np.array])
    def test_array_like(self, universe, array_like):
        array = array_like([0, 2])
        frames = [universe.trajectory.ts.frame
                  for ts in universe.trajectory[array]]
        assert_equal(frames, array)

    def test_list_indices(self, universe):
        indices = [0, 1, 2, 1, 2, 2, 0]
        frames = [universe.trajectory.ts.frame
                  for ts in universe.trajectory[indices]]
        assert_equal(frames, indices)

    @pytest.mark.parametrize('group, attr', (('position', 'positions'),
                                             ('velocity', 'velocities'),
                                             ('force', 'forces')))
    def test_no_group(self, h5md_file, outfile, attr, group):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                del g[f'particles/trajectory/{group}']
        u = mda.Universe(TPR_xvf, outfile)
        with pytest.raises(NoDataError,
                           match="This Timestep has no"):
            getattr(u.trajectory.ts, attr)

    @pytest.mark.parametrize('dset',
        ('position/value', 'position/time', 'velocity/value', 'force/value'))
    def test_unknown_unit(self, h5md_file, outfile, dset):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                g['particles'
                  '/trajectory'
                  f'/{dset}'].attrs['unit'] = 'random string'
        with pytest.raises(RuntimeError,
                           match=" is not recognized by H5MDReader."):
            u = mda.Universe(TPR_xvf, outfile)

    def test_length_unit_from_box(self, h5md_file, universe, outfile):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                del g['particles/trajectory/position']
        ref_u = universe
        uw = mda.Universe(TPR_xvf, outfile)
        assert_equal(ref_u.trajectory.units['length'],
                     uw.trajectory.units['length'])
        for ref_ts, new_ts in zip(ref_u.trajectory, uw.trajectory):
            assert_equal(ref_ts.dimensions, new_ts.dimensions)
            assert_equal(ref_ts.triclinic_dimensions,
                         new_ts.triclinic_dimensions)

    @pytest.mark.parametrize('group', ('position', 'velocity', 'force'))
    def test_changing_n_atoms(self, h5md_file, outfile, group):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                g[f'particles/trajectory/{group}/value'].resize((3, 10000, 3))
        with pytest.raises(ValueError,
                           match=" of either the postion, velocity, or force"):
            u = mda.Universe(TPR_xvf, outfile)

    def test_2D_box(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                new_box = np.ones(shape=(3, 2, 2))
                g['particles/trajectory/box'].attrs['dimension'] = 2
                del g['particles/trajectory/box/edges/value']
                g['particles/trajectory'
                  '/box/edges'].create_dataset('value', data=new_box)
        with pytest.raises(ValueError,
                           match="MDAnalysis only supports 3-dimensional"):
            u = mda.Universe(TPR_xvf, outfile)

    def test_no_box(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                del g['particles/trajectory/box/edges']
        u = mda.Universe(TPR_xvf, outfile)
        assert_equal(u.trajectory.ts.dimensions, None)

    def test_no_groups(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                del g['particles/trajectory/position']
                del g['particles/trajectory/velocity']
                del g['particles/trajectory/force']
        with pytest.raises(NoDataError,
                           match="Provide at least a position, velocity"):
            u = mda.Universe(TPR_xvf, outfile)

    def test_no_convert_units(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                groups = ['position', 'velocity', 'force']
                for name in groups:
                    del g['particles/trajectory'][name]['value'].attrs['unit']
                del g['particles/trajectory/position/time'].attrs['unit']
        u = mda.Universe(TPR_xvf, outfile, convert_units=False)
        for unit in u.trajectory.units:
            assert_equal(u.trajectory.units[unit], None)

    def test_no_units_but_convert_units_true_error(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                groups = ['position', 'velocity', 'force']
                for name in groups:
                    del g['particles/trajectory'][name]['value'].attrs['unit']
                del g['particles/trajectory/position/time'].attrs['unit']
                del g['particles/trajectory/box/edges/value'].attrs['unit']
        with pytest.raises(ValueError,
                           match="H5MD file must have readable units if"):
            u = mda.Universe(TPR_xvf, outfile, convert_units=True)

    @pytest.mark.xfail(reason='Issue #2884')
    def test_open_filestream(self, h5md_file, universe):
        with h5md_file as f:
            u = mda.Universe(TPR_xvf, h5md_file)
            for ts1, ts2 in zip(universe.trajectory, u.trajectory):
                assert_equal(ts1.positions, ts2.positions)
                assert_equal(ts1.velocities, ts2.velocities)
                assert_equal(ts1.forces, ts2.forces)

    def test_wrong_driver(self):
        with pytest.raises(ValueError,
                           match="If MPI communicator object is used to open"):
            u = mda.Universe(TPR_xvf, H5MD_xvf,
                             driver='wrong_driver',
                             comm="mock MPI.COMM_WORLD")

    def test_open_with_driver(self):
        u = mda.Universe(TPR_xvf, H5MD_xvf, driver="core")
        assert_equal(u.trajectory._file.driver, "core")

    @pytest.mark.parametrize('group1, group2', (('velocity', 'force'),
                                                ('position', 'force'),
                                                ('position', 'velocity')))
    def test_parse_n_atoms(self, h5md_file, outfile, group1, group2):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                traj_group = g['particles/trajectory']
                del traj_group[group1]
                del traj_group[group2]
                for dset in ('position/value', 'velocity/value',
                             'force/value'):
                    try:
                        n_atoms_in_dset = traj_group[dset].shape[1]
                        break
                    except KeyError:
                        continue

        u = mda.Universe(outfile)
        assert_equal(u.atoms.n_atoms, n_atoms_in_dset)

    def test_parse_n_atoms_error(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, 'w') as g:
                f.copy(source='particles', dest=g)
                f.copy(source='h5md', dest=g)
                traj_group = g['particles/trajectory']
                del traj_group['position']
                del traj_group['velocity']
                del traj_group['force']

        errmsg = "Could not construct minimal topology"
        with pytest.raises(ValueError, match=errmsg):
            u = mda.Universe(outfile)


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class TestH5MDWriterWithRealTrajectory(object):

    prec = 3

    @pytest.fixture()
    def universe(self):
        return mda.Universe(TPR_xvf, H5MD_xvf)

    @pytest.fixture()
    def universe_no_units(self):
        u = mda.Universe(TPR_xvf, H5MD_xvf, convert_units=False)
        u.trajectory.units['time'] = None
        u.trajectory.units['length'] = None
        u.trajectory.units['velocity'] = None
        u.trajectory.units['force'] = None
        return u

    @pytest.fixture()
    def Writer(self):
        return mda.coordinates.H5MD.H5MDWriter

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir) + 'h5md-writer-test.h5md'

    @pytest.fixture()
    def outtop(self, tmpdir):
        return str(tmpdir) + 'h5md-writer-top.pdb'

    @pytest.mark.parametrize('scalar, error, match',
        ((0, ValueError, "H5MDWriter: no atoms in output trajectory"),
        (0.5, IOError, "H5MDWriter: Timestep does not have")))
    def test_n_atoms_errors(self, universe, Writer, outfile,
                            scalar, error, match):
        n_atoms = universe.atoms.n_atoms * scalar
        with pytest.raises(error, match=match):
            with Writer(outfile, n_atoms) as W:
                W.write(universe)

    def test_chunk_error(self, universe, Writer, outfile):
        n_atoms = universe.atoms.n_atoms
        err = "H5MDWriter must know how many frames will be "
        with pytest.raises(ValueError, match=err):
            with Writer(outfile, n_atoms, chunks=False) as W:
                for ts in universe.trajectory:
                    W.write(universe)

    @pytest.mark.parametrize('dimensions', (None, 0))
    def test_no_dimensions(self, universe, Writer, outfile, dimensions):
        with Writer(outfile, universe.atoms.n_atoms) as W:
            for ts in universe.trajectory:
                ts.dimensions = dimensions
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        box = uw.trajectory._particle_group['box']
        assert 'edges' not in box
        assert_equal(3*['none'], box.attrs['boundary'])
        assert_equal(3, box.attrs['dimension'])

    def test_step_not_monotonic(self, universe, Writer, outfile):
        with pytest.raises(ValueError,
                           match="The H5MD standard dictates that the step "):
            with Writer(outfile, universe.atoms.n_atoms) as W:
                for ts in universe.trajectory[[0, 1, 2, 1]]:
                    W.write(universe)

        with pytest.raises(ValueError,
                           match="The H5MD standard dictates that the step "):
            with Writer(outfile, universe.atoms.n_atoms) as W:
                for ts in universe.trajectory:
                    if ts.frame == 2:
                        ts.data['step'] = 0
                    W.write(universe)

    def test_step_from_frame(self, universe, Writer, outfile):
        with Writer(outfile, universe.atoms.n_atoms) as W:
            for ts in universe.trajectory:
                del ts.data['step']
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        steps = [ts.data['step'] for ts in uw.trajectory]
        frames = [ts.frame for ts in universe.trajectory]
        for step, frame in zip(steps, frames):
            assert_equal(step, frame)

    def test_has_property(self, universe, Writer, outfile):
        with Writer(outfile, universe.atoms.n_atoms) as W:
            W.write(universe)
            # make sure property is pulled from _has dict
            assert W.has_positions == W._has['position']
            assert W.has_velocities == W._has['velocity']
            assert W.has_forces == W._has['force']
            # make sure the values are correct
            assert W.has_positions is True
            assert W.has_velocities is True
            assert W.has_forces is True

    @pytest.mark.parametrize('pos, vel, force', (
            (True, False, False),
            (True, True, False),
            (True, False, True),
            (True, True, True),
            (False, True, True),
            (False, False, True),
            (False, False, False)))
    def test_write_trajectory(self, universe, Writer, outfile,
                              pos, vel, force):
        try:
            with Writer(outfile,
                        universe.atoms.n_atoms,
                        positions=pos,
                        velocities=vel,
                        forces=force,
                        author='My Name',
                        author_email='my_email@asu.edu') as W:
                for ts in universe.trajectory:
                    W.write(universe)

            uw = mda.Universe(TPR_xvf, outfile)

            # check the trajectory contents match reference universes
            for ts, ref_ts in zip(uw.trajectory, universe.trajectory):
                assert_almost_equal(ts.dimensions, ref_ts.dimensions, self.prec)
                if pos:
                    assert_almost_equal(ts._pos, ref_ts._pos, self.prec)
                else:
                    with pytest.raises(NoDataError,
                                       match="This Timestep has no"):
                        getattr(ts, 'positions')
                if vel:
                    assert_almost_equal(ts._velocities, ref_ts._velocities,
                                        self.prec)
                else:
                    with pytest.raises(NoDataError,
                                       match="This Timestep has no"):
                        getattr(ts, 'velocities')
                if force:
                    assert_almost_equal(ts._forces, ref_ts._forces, self.prec)
                else:
                    with pytest.raises(NoDataError,
                                       match="This Timestep has no"):
                        getattr(ts, 'forces')

        # when (False, False, False)
        except(ValueError):
            with pytest.raises(ValueError,
                               match="At least one of positions, velocities"):
                with Writer(outfile,
                            universe.atoms.n_atoms,
                            positions=pos,
                            velocities=vel,
                            forces=force,
                            author='My Name',
                            author_email='my_email@asu.edu') as W:
                    for ts in universe.trajectory:
                        W.write(universe)

    def test_write_AtomGroup_with(self, universe, outfile, outtop, Writer):
        """test to write H5MD from AtomGroup"""
        ca = universe.select_atoms("protein and name CA")
        ca.write(outtop)
        with Writer(outfile, n_atoms=ca.n_atoms) as W:
            for ts in universe.trajectory:
                W.write(ca)

        uw = mda.Universe(outtop, outfile)
        caw = uw.atoms

        for orig_ts, written_ts in zip(universe.trajectory,
                                       uw.trajectory):
            assert_almost_equal(ca.positions, caw.positions, self.prec)
            assert_almost_equal(orig_ts.time, written_ts.time, self.prec)
            assert_almost_equal(written_ts.dimensions,
                                orig_ts.dimensions,
                                self.prec)

    @pytest.mark.parametrize('frames, n_frames', ((None, 1),
                                                  ('all', 3)))
    def test_ag_write(self, universe, outfile, outtop,
                      Writer, frames, n_frames):
        """test to write with ag.write()"""
        ca = universe.select_atoms("protein and name CA")
        ca.write(outtop)

        ca.write(outfile, frames=frames, format='h5md')

        uw = mda.Universe(outtop, outfile)
        caw = uw.atoms

        assert_equal(n_frames, len(uw.trajectory))
        for orig_ts, written_ts in zip(universe.trajectory,
                                       uw.trajectory):
            assert_almost_equal(ca.positions, caw.positions, self.prec)
            assert_almost_equal(orig_ts.time, written_ts.time, self.prec)
            assert_almost_equal(written_ts.dimensions,
                                orig_ts.dimensions,
                                self.prec)

    @pytest.mark.parametrize('timeunit, lengthunit, velocityunit, forceunit', (
        ('fs', 'Angstrom', 'Angstrom/ps', 'kJ/(mol*Angstrom)'),
        ('s', 'pm', 'm/s', 'Newton'),
        ('ps', 'fm', 'Angstrom/fs', 'kcal/(mol*Angstrom)',),
        ('AKMA', 'nm', 'Angstrom/AKMA', 'kcal/(mol*Angstrom)')))
    def test_write_custom_units(self, universe, outfile, Writer,
                                timeunit, lengthunit,
                                velocityunit, forceunit):
        with Writer(outfile,
                    universe.atoms.n_atoms,
                    lengthunit=lengthunit,
                    velocityunit=velocityunit,
                    forceunit=forceunit,
                    timeunit=timeunit) as W:
            for ts in universe.trajectory:
                W.write(universe)

        u = mda.Universe(TPR_xvf, outfile)
        for u_unit, custom_unit in zip(u.trajectory.units.values(),
                                       (timeunit, lengthunit,
                                       velocityunit, forceunit)):
            assert_equal(u_unit, custom_unit)

    @pytest.mark.parametrize('timeunit, lengthunit, velocityunit, forceunit', (
        ('imaginary time', None, None, None),
        (None, None, 'c', None),
        (None, None, None, 'HUGE FORCE',),
        (None, 'lightyear', None, None),))
    def test_write_bad_units(self, universe, outfile, Writer,
                             timeunit, lengthunit,
                             velocityunit, forceunit):
        with pytest.raises(ValueError, match=" is not a unit recognized by"):
            with Writer(outfile,
                        universe.atoms.n_atoms,
                        lengthunit=lengthunit,
                        velocityunit=velocityunit,
                        forceunit=forceunit,
                        timeunit=timeunit) as W:
                for ts in universe.trajectory:
                    W.write(universe)

    def test_no_units_w_convert_true(self, universe_no_units, outfile, Writer):
        # no units + convert_units = ValueError
        with pytest.raises(ValueError, match="The trajectory has no units,"):
            with Writer(outfile,
                        universe_no_units.atoms.n_atoms) as W:
                for ts in universe_no_units.trajectory:
                    W.write(universe_no_units)

    def test_no_units_w_convert_false(self, universe_no_units,
                                      outfile, Writer):
        with Writer(outfile,
                    universe_no_units.atoms.n_atoms,
                    convert_units=False) as W:
            for ts in universe_no_units.trajectory:
                W.write(universe_no_units)

            uw = mda.Universe(TPR_xvf, outfile, convert_units=False)
            for unit in uw.trajectory.units.values():
                assert_equal(unit, None)

    @pytest.mark.parametrize('convert_units', (True, False))
    def test_convert_units(self, universe, outfile, Writer,
                           convert_units):
        with Writer(outfile,
                    universe.atoms.n_atoms,
                    convert_units=convert_units) as W:
            for ts in universe.trajectory:
                W.write(universe)

        ref_units = universe.trajectory.units.items()
        uw = mda.Universe(TPR_xvf, outfile)
        uw_units = uw.trajectory.units.items()
        for u1, u2 in zip(ref_units, uw_units):
            assert_equal(u1, u2)

    @pytest.mark.parametrize('chunks', ((3, 1000, 1),
                                        (1, 1000, 3),
                                        (100, 100, 3)))
    def test_write_chunks(self, universe, outfile, Writer, chunks):
        with Writer(outfile,
                    universe.atoms.n_atoms,
                    chunks=chunks) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        for dset in (uw.trajectory._particle_group['position/value'],
                     uw.trajectory._particle_group['velocity/value'],
                     uw.trajectory._particle_group['force/value']):
            assert_equal(dset.chunks, chunks)

        for ts1, ts2 in zip(universe.trajectory, uw.trajectory):
            assert_equal(ts1.positions, ts2.positions)
            assert_equal(ts1.velocities, ts2.velocities)
            assert_equal(ts1.forces, ts2.forces)

    def test_write_chunks_with_nframes(self, universe, outfile, Writer):
        n_atoms = universe.atoms.n_atoms
        n_frames = universe.trajectory.n_frames
        with Writer(outfile,
                    n_atoms=n_atoms,
                    n_frames=n_frames) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        for dset in (uw.trajectory._particle_group['position/value'],
                     uw.trajectory._particle_group['velocity/value'],
                     uw.trajectory._particle_group['force/value']):
            assert_equal(dset.chunks, (1, n_atoms, 3))

        for ts1, ts2 in zip(universe.trajectory, uw.trajectory):
            assert_equal(ts1.positions, ts2.positions)
            assert_equal(ts1.velocities, ts2.velocities)
            assert_equal(ts1.forces, ts2.forces)

    def test_write_contiguous1(self, universe, Writer, outfile):
        n_atoms = universe.atoms.n_atoms
        n_frames = len(universe.trajectory)
        with Writer(outfile,
                    n_atoms=n_atoms,
                    n_frames=n_frames,
                    chunks=False) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        for dset in (uw.trajectory._particle_group['position/value'],
                     uw.trajectory._particle_group['velocity/value'],
                     uw.trajectory._particle_group['force/value']):
            assert_equal(dset.chunks, None)

    def test_write_contiguous2(self, universe, Writer, outfile):
        ag = universe.select_atoms('all')
        n_frames = len(ag.universe.trajectory)
        ag.write(outfile, frames='all', n_frames=n_frames, chunks=False)

        uw = mda.Universe(TPR_xvf, outfile)
        for dset in (uw.trajectory._particle_group['position/value'],
                     uw.trajectory._particle_group['velocity/value'],
                     uw.trajectory._particle_group['force/value']):
            assert_equal(dset.chunks, None)

    @pytest.mark.parametrize('filter, opts', (('gzip', 1),
                                              ('gzip', 9),
                                              ('lzf', None)))
    def test_write_with_compression(self, universe,
                                    outfile, Writer,
                                    filter, opts):
        with Writer(outfile,
                    universe.atoms.n_atoms,
                    compression=filter,
                    compression_opts=opts) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        dset = uw.trajectory._particle_group['position/value']
        assert_equal(dset.compression, filter)
        assert_equal(dset.compression_opts, opts)

    @pytest.mark.xfail(os.name == 'nt',
                       reason="occasional PermissionError on windows")
    @pytest.mark.parametrize('driver', ('core', 'stdio'))
    def test_write_with_drivers(self, universe, outfile, Writer, driver):
        with Writer(outfile,
                    universe.atoms.n_atoms,
                    driver=driver) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile, driver=driver)
        file = uw.trajectory._file
        assert_equal(file.driver, driver)

    def test_parallel_disabled(self, universe, Writer, outfile,
                               driver='mpio'):
        with pytest.raises(ValueError,
                           match="H5MDWriter: parallel writing with MPI I/O "):
            with Writer(outfile,
                        universe.atoms.n_atoms,
                        driver=driver) as W:
                for ts in universe.trajectory:
                    W.write(universe)

    def test_timestep_not_modified_by_writer(self, universe, Writer, outfile):
        trj = universe.trajectory
        ts = trj.ts

        trj[-1]  # last timestep (so that time != 0)
        x = ts._pos.copy()
        time = ts.time

        with Writer(outfile, trj.n_atoms) as W:
            # last timestep (so that time != 0) (say it again, just in case...)
            trj[-1]
            W.write(universe)

        assert_equal(
            ts._pos,
            x,
            err_msg="Positions in Timestep were modified by writer.")
        assert_equal(
            ts.time, time, err_msg="Time in Timestep was modified by writer.")


class TestH5PYNotInstalled(object):
    """Tests RuntimeErrors when h5py not installed"""

    @pytest.fixture(autouse=True)
    def block_h5py(self, monkeypatch):
        monkeypatch.setattr(
            sys.modules['MDAnalysis.coordinates.H5MD'], 'HAS_H5PY', False)

    @pytest.fixture()
    def Writer(self):
        return mda.coordinates.H5MD.H5MDWriter

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir) + 'h5md-writer-test.h5md'

    def test_reader_no_h5py(self):
        with pytest.raises(RuntimeError, match="Please install h5py"):
            u = mda.Universe(TPR_xvf, H5MD_xvf)

    def test_writer_no_h5py(self, Writer, outfile):
        u = mda.Universe(TPR_xvf, TRR_xvf)
        with pytest.raises(RuntimeError,
                           match="H5MDWriter: Please install h5py"):
            with Writer(outfile,
                        u.atoms.n_atoms) as W:
                for ts in u.trajectory:
                    W.write(universe)
